"""Implementation of XBAmodel class.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import os
import time
import numpy as np
import sbmlxdf
from scipy.sparse import coo_array

from .xba_unit_def import XbaUnitDef
from .xba_compartment import XbaCompartment
from .xba_function import XbaFunction
from .xba_parameter import XbaParameter
from .xba_reaction import XbaReaction
from .xba_species import XbaSpecies
from .fbc_objective import FbcObjective
from .fbc_gene_product import FbcGeneProduct


# TODO: implement isGBAmodel(), isFBAmodel(), isRBAmodel()
class XbaModel:

    def __init__(self, sbml_file):
        if os.path.exists(sbml_file) is False:
            print(f'{sbml_file} does not exist')
            raise FileNotFoundError

        print(f'loading: {sbml_file} (last modified: {time.ctime(os.path.getmtime(sbml_file))})')
        sbml_model = sbmlxdf.Model(sbml_file)
        if sbml_model.isModel is False:
            print(f'{sbml_file} seems not to be a valid SBML model')
            return
        model_dict = sbml_model.to_df()
        self.unit_defs = {udid: XbaUnitDef(row)
                          for udid, row in model_dict['unitDefs'].iterrows()}
        self.compartments = {cid: XbaCompartment(row)
                             for cid, row in model_dict['compartments'].iterrows()}
        self.parameters = {pid: XbaParameter(row)
                           for pid, row in model_dict['parameters'].iterrows()}
        if 'funcDefs' in model_dict:
            self.functions = {fid: XbaFunction(row)
                              for fid, row in model_dict['funcDefs'].iterrows()}
        else:
            self.functions = {}

        self.species = {sid: XbaSpecies(row)
                        for sid, row in model_dict['species'].iterrows()}
        self.reactions = {rid: XbaReaction(row, self.species, self.functions, self.compartments)
                          for rid, row in model_dict['reactions'].iterrows()}

        for r in self.reactions.values():
            r.set_enzyme(self.species)
            r.add_species_usage(self.species)

        if 'fbcObjectives' in model_dict:
            self.fbc_objectives = {oid: FbcObjective(row)
                                   for oid, row in model_dict['fbcObjectives'].iterrows()}
        if 'fbcGeneProducts' in model_dict:
            self.fbc_gene_products = {gp_id: FbcGeneProduct(row)
                                      for gp_id, row in model_dict['fbcGeneProducts'].iterrows()}

    def get_stoic_matrix(self, sids, rids):
        """retrieve stoichiometric sub-matrix [sids x rids].

        :param sids: species ids - rows of sub-matrix
        :type sids: list of strings
        :param rids: reaction ids - columns of sub-matrix
        :type rids: list of strings
        :returns: stoic_mat, stoichmetric sub-matrix
        :rtype: np.ndarray
        """
        stoic_mat = np.zeros((len(sids), len(rids)))
        map_metab_row = {sid: idx for idx, sid in enumerate(sids)}
        for col, rid in enumerate(rids):
            for sid, stoic in self.reactions[rid].reactants.items():
                if sid in map_metab_row:
                    stoic_mat[map_metab_row[sid], col] = -stoic
            for sid, stoic in self.reactions[rid].products.items():
                if sid in map_metab_row:
                    stoic_mat[map_metab_row[sid], col] = stoic
        return stoic_mat

    def get_stoic_matrix_coo(self, sids, rids):
        """retrieve stoichiometric sub-matrix [sids x rids].

        :param sids: species ids - rows of sub-matrix
        :type sids: list of strings
        :param rids: reaction ids - columns of sub-matrix
        :type rids: list of strings
        :returns: stoichmetric sub-matrix in sparse format
        :rtype: scipy.sparse.coo_array
        """
        sid2idx = {sid: idx for idx, sid in enumerate(sids)}
        stoic_data = []
        for col_idx, rid in enumerate(rids):
            for sid, stoic in self.reactions[rid].reactants.items():
                if sid in sids:
                    stoic_data.append([sid2idx[sid], col_idx, -stoic])
            for sid, stoic in self.reactions[rid].products.items():
                if sid in sids:
                    stoic_data.append([sid2idx[sid], col_idx, stoic])
        coo_data = np.array(stoic_data)
        stoic_mat_coo = coo_array((coo_data[:, 2], coo_data[:, :2].T), shape=(len(sids), len(rids)))
        return stoic_mat_coo
