"""Implementation of GBA Problem class.

   - several process machines supported (yet to be validated)
   - non-enzymatic metabolic reactions supported (yet to be validated)

   Yet to implement:
   - enzyme degradation
   - enzyme transitions
   - several compartments constraints
   - target concentrations and target fluxes

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np
import pandas as pd
from scipy.sparse import coo_array, hstack, vstack

from xbanalysis.solvers.glpk_linear_problem import GlpkLinearProblem


class RbaProblem:

    def __init__(self, xba_model):
        """

        :param xba_model:
        """

        self.xba_model = xba_model

        # metabolic reactions and enzyme transitions (no FBA reactions, no protein synthesis, no degradation)
        self.rids = np.array([r.id for r in xba_model.reactions.values()
                              if r.sboterm.is_in_branch('SBO:0000167') is True and
                              r.sboterm.is_in_branch('SBO:0000179') is False])
        self.rid2enz = {rid: xba_model.reactions[rid].enzyme for rid in self.rids}
        self.enz_sids = np.array([enz_sid for enz_sid in self.rid2enz.values() if enz_sid is not None])
        self.enz_sids_rev = np.array([enz_sid for rid, enz_sid in self.rid2enz.items()
                                      if enz_sid is not None and  xba_model.reactions[rid].reversible])
        self.enz2idx = {enz: idx for idx, enz in enumerate(self.enz_sids)}
        self.processes = {s.rba_process_machine: sid for sid, s in xba_model.species.items()
                          if hasattr(s, 'rba_process_machine')}
        self.densities = {cid: c.rba_frac_gcdw for cid, c in xba_model.compartments.items()
                          if hasattr(c, 'rba_frac_gcdw')}

        # metabolites and enzymes used in metabolic reactions and enzyme transitions
        used_sids = set()
        for rid in self.rids:
            used_sids |= set(xba_model.reactions[rid].reactants)
            used_sids |= set(xba_model.reactions[rid].products)
        self.sids = np.array([sid for sid in used_sids if xba_model.species[sid].constant is False])
        self.sid2idx = {sid: idx for idx, sid in enumerate(self.sids)}
        self.ext_conc = {sid: xba_model.species[sid].initial_conc for sid in used_sids
                         if xba_model.species[sid].constant is True}

        self.n_rid = len(self.rids)
        self.n_enz = len(self.enz_sids)
        self.n_pm = len(self.processes)

        self.rba_cols = list(self.rids) + list(self.enz_sids) + list(self.processes)
        self.rba_rows = (list(self.sids) + list(self.processes)
                         + [sid + '_feff' for sid in self.enz_sids]
                         + [sid + '_reff' for sid in self.enz_sids_rev]
                         + list(self.densities))

        self.is_rba_model = self.check_rba_model(xba_model)
        if self.is_rba_model is False:
            print('Error: not a GBA model - missing parameters')
            return

        fix_m_coo, var_m_coo, row_bnds_m = self._construct_mass_balance()
        fix_pc_coo, var_pc_coo, row_bnds_pc = self._construct_process_capacities()
        fix_eff_coo, var_eff_coo, row_bnds_eff = self._construct_efficencies()
        fix_cd_coo, var_cd_coo, row_bnds_cd = self._construct_densities()
        self.fix_cd_coo = fix_cd_coo
        self.fix_mat_coo = vstack((fix_m_coo, fix_pc_coo, fix_eff_coo, fix_cd_coo))
        self.var_mat_coo = vstack((var_m_coo, var_pc_coo, var_eff_coo, var_cd_coo))
        self.fix_row_bnds = np.vstack((np.zeros_like(row_bnds_m), np.zeros_like(row_bnds_pc),
                                       row_bnds_eff, row_bnds_cd))
        self.var_row_bnds = np.vstack((row_bnds_m, row_bnds_pc, np.zeros_like(row_bnds_eff),
                                       np.zeros_like(row_bnds_cd)))

        # in RBA we only check feasibility.
        # Here we maximize enzyme amount and check that objective value e.g. > 1e-10 mmol/gcdw
        self.obj_dir = 'maximize'
        self.obj_coefs = np.hstack((np.zeros(self.n_rid), np.ones(self.n_enz + self.n_pm)))
        col_lbs = np.hstack(([xba_model.reactions[rid].fbc_lb for rid in self.rids],
                             np.zeros(self.n_enz + self.n_pm)))
        col_ubs = np.hstack(([xba_model.reactions[rid].fbc_ub for rid in self.rids],
                             np.nan * np.ones(self.n_enz + self.n_pm)))
        self.col_bnds = np.vstack((col_lbs, col_ubs))
        self.scaling = '2N'
        self.max_iter = 20

    def check_rba_model(self, xba_model):
        """Check if rba related data available in the XBA model

        :param xba_model: Xba model with data extracted from SBML file
        :type xba_model: XbaModel
        :return: indicator if model contains data required for RBA problem formulation
        :rtype: bool
        """
        is_rba_model = True

        r_check = {'flux lower bound': 'fbc_lb', 'flux upper bound': 'fbc_ub', 'kapp forward': 'rba_kapp_f'}
        for check, param in r_check.items():
            missing = [rid for rid in self.rids if hasattr(xba_model.reactions[rid], param) is False]
            if len(missing) > 0:
                print(f'missing {check} in:', missing)
                is_rba_model = False

        missing = [rid for rid in self.rids if xba_model.reactions[rid].reversible
                   and hasattr(xba_model.reactions[rid], 'rba_kapp_r') is False]
        if len(missing) > 0:
            print('missing kapp reverse in:', missing)
            is_rba_model = False

        missing = [sid for sid in list(self.sids) + list(self.enz_sids) + list(self.processes.values())
                   if hasattr(xba_model.species[sid], 'mw') is False]
        if len(missing) > 0:
            print('missing molecular weight in:', missing)
            is_rba_model = False

        missing = [sid for sid in list(self.processes.values())
                   if hasattr(xba_model.species[sid], 'rba_process_capacity') is False]
        if len(missing) > 0:
            print('missing process capacity in:', missing)
            is_rba_model = False

        if len(self.densities) == 0:
            print('missing a density constraint')
            is_rba_model = False
        return is_rba_model

    def check_growth_rate_feasibility(self, gr, short_result=True):
        """Check, if a selected growth rate is supported by the model

        create a Linear Programming problem and check its feasibility

        :param gr: current growth rate
        :type gr: float
        :param short_result: short result (only objective value and status) or long result
        :type short_result: bool (optional, default: False)
        :return: result of optimization
        :rtype: dict
        """
        constr_coefs_coo = coo_array(self.fix_mat_coo + gr * self.var_mat_coo)

        lp = GlpkLinearProblem(constr_coefs_coo, self.fix_row_bnds, self.col_bnds)
        lp.set_obj_coefs(self.obj_coefs)
        lp.set_obj_dir(self.obj_dir)
        res = lp.solve(short_result, scaling=self.scaling)
        del lp
        return res

    def rba_optimize(self):
        """Resource Balance Analysis

        According to Goelzer, Fromion:
          Goelzer et al., 2015. Quantitative prediction of genome-wide resource allocation
          in bacteria. Metab Eng, 32, 232-243

        Stage 1: course range finding of supported growth rate
        Stage 2: zooming in on supported growth rate

        :return: result including maximal growth rate
        :rtype: dict
        """
        if self.is_rba_model is False:
            print('Error: not a RBA model - missing parameters')
            res = {'success': False, 'fun': 0.0, 'message': 'not an RBA model'}
            return res

        # Preanalysis: find range of growth rates
        gr = {'lb': 0.0, 'ub': np.nan, 'try': 1.0}
        for i in range(self.max_iter):
            res = self.check_growth_rate_feasibility(gr['try'])
            if res['success'] and res['fun'] > 1e-10:
                gr['lb'] = gr['try']
                gr['try'] *= 10
            else:
                if gr['try'] <= 1e-4:
                    break
                gr['ub'] = gr['try']
                gr['try'] /= 10
            if gr['lb'] > 0.0 and np.isfinite(gr['ub']):
                break
        if gr['lb'] == 0.0 or np.isnan(gr['ub']):
            res = {'success': False, 'fun': 0.0, 'message': 'no solution found'}
            return res

        # find exact growth rate
        for i in range(self.max_iter):
            gr['try'] = 0.5 * (gr['lb'] + gr['ub'])
            res = self.check_growth_rate_feasibility(gr['try'])
            if res['success'] and res['fun'] > 1e-10:
                gr['lb'] = gr['try']
            else:
                gr['ub'] = gr['try']
            if (gr['ub'] - gr['lb']) < gr['try'] * 1e-5:
                break

        # get optimum values for the final solution
        gr['try'] = gr['lb']
        res = self.check_growth_rate_feasibility(gr['try'], short_result=False)
        res['fun'] = gr['try']
        return res

    def _get_efficiencies(self):
        """Determine efficiencies for all enzymes """

        efficiencies = {}
        for mr_rid, enz_sid in self.rid2enz.items():
            # check for enzyme catalyzed reactions
            if enz_sid is not None:
                r = self.xba_model.reactions[mr_rid]
                saturation = 1.0
                if hasattr(r, 'rba_km_ext'):
                    for sid in r.reactants:
                        if sid in self.ext_conc:
                            saturation *= self.ext_conc[sid] / (self.ext_conc[sid] + r.rba_km_ext)
                eff_fwd = r.rba_kapp_f * 3600.0 * saturation
                saturation = 1.0
                if r.reversible:
                    if hasattr(r, 'rba_km_ext'):
                        for sid in r.products:
                            if sid in self.ext_conc:
                                saturation *= self.ext_conc[sid] / (self.ext_conc[sid] + r.rba_km_ext)
                    eff_rev = r.rba_kapp_r * 3600.0 * saturation
                else:
                    eff_rev = 0.0
                efficiencies[enz_sid] = [eff_fwd, eff_rev]
        return efficiencies

    def _construct_mass_balance(self):

        n_sid = len(self.sids)
        psenz_rids = [self.xba_model.species[enz_sid].ps_rid for enz_sid in self.enz_sids]
        pspm_rids = [self.xba_model.species[enz_sid].ps_rid for enz_sid in self.processes.values()]

        fix_m_x_r_coo = self.xba_model.get_stoic_matrix_coo(self.sids, self.rids)
        var_m_x_enz_pm_coo = self.xba_model.get_stoic_matrix_coo(self.sids, psenz_rids + pspm_rids)

        fix_m_coo = hstack((fix_m_x_r_coo, coo_array((n_sid, self.n_enz + self.n_pm))))
        var_m_coo = hstack((coo_array((n_sid, self.n_rid)), var_m_x_enz_pm_coo))

        targets = np.array([getattr(self.xba_model.species[sid], 'rba_target', 0.0) for sid in self.sids])
        row_bnds_m = np.vstack((targets, targets)).T
        return fix_m_coo, var_m_coo, row_bnds_m

    def _construct_process_capacities(self):

        pm2idx = {pm: idx for idx, pm in enumerate(self.processes)}

        mat_data = []
        for pm in self.processes:
            for col_idx, enz_sid in enumerate(list(self.enz_sids) + list(self.processes.values())):
                p_cost = self.xba_model.species[enz_sid].rba_process_costs.get(pm, 0.0)
                if p_cost > 0.0:
                    mat_data.append([pm2idx[pm], self.n_rid + col_idx, p_cost])

        coo_data = np.array(mat_data)
        var_pc_coo = coo_array((coo_data[:, 2], coo_data[:, :2].T),
                               shape=(self.n_pm, self.n_rid + self.n_enz + self.n_pm))
        mat_data = []
        for pm, enz_sid in self.processes.items():
            capacity = self.xba_model.species[enz_sid].rba_process_capacity * 3600.0
            mat_data.append([pm2idx[pm], self.n_rid + self.n_enz + pm2idx[pm], -capacity])
        coo_data = np.array(mat_data)
        fix_pc_coo = coo_array((coo_data[:, 2], coo_data[:, :2].T),
                               shape=(self.n_pm, self.n_rid + self.n_enz + self.n_pm))

        row_bnds_pc = np.zeros((self.n_pm, 2))
        return fix_pc_coo, var_pc_coo, row_bnds_pc

    def _construct_efficencies(self):
        """Construct enzyme efficiency constraints

        :return: fixed and variable coefficient matrices, row bounds
        """
        efficiencies = self._get_efficiencies()
        rid2idx = {rid: idx for idx, rid in enumerate(self.rids)}

        mat_data = []
        enz2rid = {enz: rid for rid, enz in self.rid2enz.items()}
        n_rows = self.n_enz + len(self.enz_sids_rev)
        # forward efficiencies (for all enzyme catalyzed reactions)
        for row_idx, enz_sid in enumerate(self.enz_sids):
            feff = efficiencies[enz_sid][0]
            mat_data.append([row_idx, rid2idx[enz2rid[enz_sid]], 1.0])
            mat_data.append([row_idx, self.n_rid + self.enz2idx[enz_sid], -feff])
        # backward efficiencies (only for reversible catalyzed reactions)
        for row_idx, enz_sid in enumerate(self.enz_sids_rev):
            reff = efficiencies[enz_sid][1]
            mat_data.append([self.n_enz + row_idx, rid2idx[enz2rid[enz_sid]], -1.0])
            mat_data.append([self.n_enz + row_idx, self.n_rid + self.enz2idx[enz_sid], -reff])
        coo_data = np.array(mat_data)
        fix_eff_coo = coo_array((coo_data[:, 2], coo_data[:, :2].T),
                                shape=(n_rows, self.n_rid + self.n_enz + self.n_pm))
        var_eff_coo = coo_array(fix_eff_coo.shape)
        # enzyme efficiency inequalities ≤ 0
        row_bnds_eff = np.vstack((np.ones(n_rows) * np.nan, np.zeros(n_rows))).T
        return fix_eff_coo, var_eff_coo, row_bnds_eff

    def _construct_densities(self):
        cid2idx = {cid: idx for idx, cid in enumerate(self.densities)}
        pm2idx = {sid: idx for idx, sid in enumerate(self.processes.values())}

        mat_data = []
        for sid in self.enz_sids:
            cid = self.xba_model.species[sid].compartment
            if cid in self.densities:
                mmol_per_l = self.xba_model.species[sid].mw / 1000
                mat_data.append([cid2idx[cid], self.n_rid + self.enz2idx[sid], mmol_per_l])
        for sid in self.processes.values():
            cid = self.xba_model.species[sid].compartment
            if cid in self.densities:
                mmol_per_l = self.xba_model.species[sid].mw / 1000
                mat_data.append([cid2idx[cid], self.n_rid + self.n_enz + pm2idx[sid], mmol_per_l])
        coo_data = np.array(mat_data)
        fix_cd_coo = coo_array((coo_data[:, 2], coo_data[:, :2].T),
                               shape=(len(cid2idx), self.n_rid + self.n_enz + self.n_pm))
        var_cd_coo = coo_array(fix_cd_coo.shape)
        # density inequality constraints: 0 ≤ constraint ≤ max_density
        row_bnds_cd = np.vstack((np.zeros(len(cid2idx)), [val for val in self.densities.values()])).T
        return fix_cd_coo, var_cd_coo, row_bnds_cd

    def process_result(self, res):
        """Helper, to convert RBA result into some DataFrames

        :param res: result from a previous RBA optimization
        :type res: dict
        :return: fluxes and enzyme information
        :rtype: two pandas DataFrames
        """

        rho = 300
        if 'rho' in self.xba_model.parameters:
            rho = self.xba_model.parameters['rho'].value

        # identify the biomass reaction name for Escher Map integration
        fba_rid = 'biomass'
        for rid in self.xba_model.reactions:
            if repr(self.xba_model.reactions[rid].sboterm) == 'SBO:0000629':
                fba_rid = rid
        # retrieve translational capacity of the ribosome
        tl_flux = 0.0
        for pm_idx, enz_sid in enumerate(self.processes.values()):
            if repr(self.xba_model.species[enz_sid].sboterm) == 'SBO:0000250':
                tl_capacity = self.xba_model.species[enz_sid].rba_process_capacity * 3600.0
                ribo_conc = res['x'][self.n_rid + self.n_enz + pm_idx]
                tl_flux = ribo_conc * tl_capacity
                break

        df_fluxes = pd.DataFrame(np.hstack((res['x'][:len(self.rids)], tl_flux)),
                                 index=list(self.rids) + [fba_rid], columns=['mmol gDW-1 h-1'])
        df_fluxes.index.name = 'reaction'

        enz_sids = list(self.enz_sids) + list(self.processes.values())
        enz_mmol_per_l = res['x'][len(self.rids):] * rho
        enz_g_per_mmol = np.sum(self.fix_cd_coo.toarray()[:, len(self.rids):], axis=0)
        enz_g_per_l = enz_mmol_per_l * enz_g_per_mmol
        df_enz = pd.DataFrame(np.vstack((enz_mmol_per_l, enz_g_per_l)).T,
                              index=enz_sids, columns=['mmol/l', 'g/l'])
        df_enz.index.name = 'enzyme'

        return df_fluxes, df_enz
